
/**  Little test / demo program for OpenGL to get you started
*
* @author
*   Gabriel Zachmann, Sep 2006
*
* @bug
*   I'm sure there is still some bugs left ....
**/

#include <QtGui>
#include <QtOpenGL>
#include <QMouseEvent>

#include <math.h>

#include "glwidget.h"


GLWidget::GLWidget(QWidget * father )
    : QGLWidget(father),
		m_xTrans(0.0), m_yTrans(0.0), m_zTrans(-5.0)
{
	m_backgroundcolor[0] = 0.0;  m_backgroundcolor[1] = 0.0;  m_backgroundcolor[2] = 0.0;
    setFocusPolicy( Qt::ClickFocus ); //needed for keyboard-input

	// initialize rotationMatrix with Idendity
	for(int i = 1; i < 4; i++) 
	{
		for(int j = 0; j < i; j++) 
		{
			m_RotationMatrix[i*4+j] = m_RotationMatrix[j*4+i]= 0.0;
		}
	}
	m_RotationMatrix[0] = m_RotationMatrix[5] = m_RotationMatrix[10] = m_RotationMatrix[15] = 1.0;
}


GLWidget::~GLWidget()
{
    makeCurrent();
}


QSize GLWidget::minimumSizeHint() const
{
    return QSize(800, 800);
}

QSize GLWidget::sizeHint() const
{
    return QSize(800, 800);
}


void GLWidget::setBackgroundColor(float v)
{
	m_backgroundcolor[0] = v;
    m_backgroundcolor[1] = v;
    m_backgroundcolor[2] = v;
	updateGL();
}


void GLWidget::initializeGL()
{
	// set colorbuffer to this color if glClear() is called
    glClearColor( m_backgroundcolor[0], m_backgroundcolor[1], m_backgroundcolor[2], 0.0 );
	// set opengl states
    glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glShadeModel(GL_FLAT); // we use flat shading
	glEnable(GL_LIGHTING); // we use lighting
	glEnable(GL_LIGHT0); // only one light, and default position  (1, 1, 1)
	glEnable(GL_NORMALIZE); // tell openGL to normalize given normal vectors at input and after transformations
}


// Warning: don't make paintGL() and resizeGL() const!
// otherwise, the rendering will produce complete garbage.
// (is this a bug in Qt? g++? opengl?)

void GLWidget::paintGL()
{
	// clear framebuffer
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	// push current matrix to stack
	glPushMatrix();

	// manage simple navigation ( rotation and translation )
	glLoadIdentity();
	glTranslatef(m_xTrans, m_yTrans, m_zTrans);
	glMultMatrixf(m_RotationMatrix); //

	drawTetrahedron();

	// restore matrix from stack
	glPopMatrix();
}


void GLWidget::resizeGL( int w, int h )		// = width & height
{
    int side = qMin(w, h);
    glViewport((w - side) / 2, (h - side) / 2, side, side);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glFrustum(-1.0, +1.0, -1.0, +1.0, 2.0, 150.0);
    glMatrixMode(GL_MODELVIEW);
}


void GLWidget::mousePressEvent( QMouseEvent * e /*the event*/ )
{
    m_lastPos = e->pos();
}


void GLWidget::keyPressEvent ( QKeyEvent * e )
{
    if ( !isValid() )
        return;

    switch ( e->key() )
    {
        case Qt::Key_B:
            if ( m_backgroundcolor[0] == 1.0 )
			{
                setBackgroundColor( 0.0 );
			} else
			{
                setBackgroundColor( 1.0 );
			}
			glClearColor( m_backgroundcolor[0], m_backgroundcolor[1], m_backgroundcolor[2], 0.0 );
			updateGL();
            break;
        default:
            return;
    }       
}


void GLWidget::mouseMoveEvent( QMouseEvent * e /*the event*/ )
{
    int dx = e->x() - m_lastPos.x();
    int dy = e->y() - m_lastPos.y();

	bool ctrl_key = e->modifiers() & Qt::MetaModifier;		// only needed for Mac OS X, but doesn't hurt on other OSes

	if ( (e->buttons() & Qt::RightButton) ||
		 ctrl_key )
	{
		m_zTrans += 0.2 * dy;
	}
	else if ( e->buttons() & Qt::MidButton )
	{
		m_xTrans += 0.01 * dx;
		m_yTrans -= 0.01 * dy;
	}
	else if (e->buttons() & Qt::LeftButton)
	{
		glPushMatrix();
		glLoadIdentity();
		glRotatef(0.5*dx,0.0, 1.0, 0.0);
		glRotatef(0.5*dy,1.0, 0.0, 0.0);
		glMultMatrixf(m_RotationMatrix);
		glGetFloatv(GL_MODELVIEW_MATRIX,m_RotationMatrix);
		glPopMatrix();
    }
    m_lastPos = e->pos();
	e->accept();
	updateGL();
}

void GLWidget::drawTetrahedron()
{
	// indices of vertices to build the 4 triangles a tetrahedron contains
	int indices[4][3] = { { 0, 1, 2}, {2, 1, 3}, {0, 3, 1}, {0, 2, 3} };

	// Vertex coordinates
	GLfloat vertices[4][3] = { { 0, 0, 0}, 
							   { 0.5, sqrt(3.0f/4.0f), 0}, 
							   { 1, 0, 0}, 
							   { 0.5, sqrt(1.0f/12.0f), sqrt(2.0f/3.0f)} };

	// calculate normals : this is inefficient, but its a very simple example.....
	GLfloat normals[4][3];
	for(int i = 0; i < 4; i++)
	{
		getNormal(normals[i],vertices[indices[i][0]],vertices[indices[i][1]],vertices[indices[i][2]]);
	}

	// material
	GLfloat diffuse[4][4] = { { 1, 0 , 0, 1}, { 0, 1, 0, 1}, { 0, 0, 1, 1}, {1, 1, 0, 1} };

	// translate center of tetrahedron into origin
	glTranslatef( -0.5, -sqrt(1.0f/12.0f), -sqrt(1.0f/24.0f) );

	// draw it
	glBegin(GL_TRIANGLES);
	{
		for(int i = 0; i < 4; i++)
		{
			glMaterialfv(GL_FRONT,GL_DIFFUSE,diffuse[i]);
			glNormal3fv(normals[i]);
			glVertex3fv(vertices[indices[i][0]]);
			glVertex3fv(vertices[indices[i][1]]);
			glVertex3fv(vertices[indices[i][2]]);
		}
	}
	glEnd();
}

// calculates normal for a triangle v0-v1-v2, with vi is a 3-dim Vector
void GLWidget::getNormal(GLfloat normal[3], const GLfloat v0[3], const GLfloat v1[3], const GLfloat v2[3])
{ 	
  GLfloat d10[3], d21[3];
  for(int i = 0; i < 3; i++)
  {
	d10[i] = v1[i] - v0[i];
	d21[i] = v2[i] - v1[i];
  }
  normal[0] = d10[1]*d21[2] - d10[2]*d21[1];
  normal[1] = d10[2]*d21[0] - d10[0]*d21[2];
  normal[2] = d10[0]*d21[1] - d10[1]*d21[0];
}
